import os
DICTIONARY_BUCKET_URL = "https://s3.amazonaws.com/ieml-dictionary-versions/"
DICTIONARY_VERSIONS_FOLDER = "data/dictionary/"

DICTIONARY_DEFAULT_VERSION = '2017-06-07_00:00:00'
DICTIONARY_FOLDER = os.path.join(os.path.dirname(__file__), "data/dictionary")

AWS_ACCESS_KEY_ID = ''
AWS_SECRET_ACCESS_KEY = ''
